/**
 *This file provides the interface of the Density Packet Header.
 */

#ifndef __node_num__
#define __node_num__
const int NODENUM=60;
#endif
  
#ifndef __density_packet__
#define __density_packet__
#include <sys/types.h>
#include "common/packet.h"

/* =====================================================================
   Packet Formats...
   ===================================================================== */
#define DENSITY_DISC  	    0x01
#define DENSITY_DST_REQ   	0x04
#define DENSITY_DST_REP   	0x08
#define DENSITY_DELEG_CALC  0x10
#define DENSITY_HELLO       0x12
#define PING                0x14

/*
 * Density Protocol Header Macros
 */
#define HDR_DEN(p)		((struct hdr_den*)hdr_den::access(p))
#define HDR_DEN_DISC(p)  	((struct hdr_den_disc*)hdr_den::access(p))
#define HDR_DEN_DST_REQ(p)	((struct hdr_den_dst_req*)hdr_den::access(p))
#define HDR_DEN_DST_REP(p)	((struct hdr_den_dst_rep*)hdr_den::access(p))
#define HDR_DEN_DELEG_CALC(p)	((struct hdr_den_deleg_calc*)hdr_den::access(p))
#define HDR_DEN_HELLO(p)	((struct hdr_den_hello*)hdr_den::access(p))
#define HDR_PING(p)	((struct hdr_ping*)hdr_den::access(p))

/*
 *General Density Header - shared by all
 */
struct hdr_den {
        u_int8_t        type;
        u_int32_t       srcid;
	
	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_den* access(const Packet* p) {
		return (hdr_den*) p->access(offset_);
	}
};
                

 
/*
 *Boundary Identification protocol: discovery packet.
 *This packet is used by nodes to advertise their presence.
 */
  
struct hdr_den_disc {
        u_int8_t        disc_type;
        u_int32_t       disc_srcid;
};     

/*Election protocol: distance request packet*/
struct hdr_den_dst_req {
        u_int8_t        dist_req_type;
        u_int32_t       dist_req_srcid;
        u_int8_t        dist_req_nHops;
        u_int32_t       dist_req_IN;
};

/*Election protocol: distance reply packet*/
struct hdr_den_dst_rep {
        u_int8_t        dist_rep_type;
        u_int32_t       dist_rep_srcid;
        u_int8_t        dist_rep_nHops;
        u_int32_t       dist_rep_IN;
        u_int32_t        dist_rep_reply_fw;

};

/*Election protocol: next IN delegation packet*/
struct hdr_den_deleg_calc {
  
        u_int8_t        deleg_type;
        u_int32_t       deleg_srcid;
        
        /*The CURRENT WORST value of a node*/
        u_int8_t        worst_val;
	      /*The CURRENT BEST value of a node*/
        u_int8_t        best_val;
	      /*The CURRENT BEST NODE*/
	      u_int32_t        best_node;
        /*The list of unexplored nodes*/
	      int       unexplored[NODENUM];
	      /*The number of unexplored nodes in the unexplored list*/
	      u_int8_t        unexplored_num;
	      /*The list of explored nodes*/
	      int      explored[NODENUM];
        /*The number of explored nodes*/
	      u_int8_t        explored_num;
        /*The number of solutions equals to current best explored*/
	      u_int8_t        best_num;

};

/*Boundary maintenance protocol: Hello packets advertising nodes presence*/
struct hdr_den_hello {

        u_int8_t        hello_type;
        u_int32_t       hello_srcid;

};

/*PING Test packet*/
struct hdr_ping {
        u_int8_t        type;
        u_int32_t       srcid;
};

/*For the calculation of header bytes number*/
union hdr_all_den {
        hdr_den                 hdrden;
        hdr_den_disc            disc;
        hdr_den_dst_req         dstreq;
        hdr_den_dst_req         dstrep;
        hdr_den_deleg_calc      delegcalc;
        hdr_den_hello           hello;
        hdr_ping                ping;
};

        
#endif
